@@ -1,7 +1,9 @@
 # -*- coding: utf-8 -*-
 from __future__ import absolute_import
 import re
+from collections import defaultdict
 from functools import partial
+from adblockparser.utils import split_data
 
 
 class AdblockRule(object):
@@ -72,6 +74,10 @@ class AdblockRule(object):
     OPTIONS_SPLIT_PAT = ',(?=~?(?:%s))' % ('|'.join(BINARY_OPTIONS + ["domain"]))
     OPTIONS_SPLIT_RE = re.compile(OPTIONS_SPLIT_PAT)
 
+    __slots__ = ['raw_rule_text', 'is_comment', 'is_html_rule', 'is_exception',
+                 'raw_options', 'options', '_options_keys', 'rule_text',
+                 'regex']
+
     def __init__(self, rule_text):
         self.raw_rule_text = rule_text
 
@@ -104,27 +110,27 @@ class AdblockRule(object):
         else:
             self.regex = self.rule_to_regex(rule_text)
 
-    def match_url(self, url, params=None):
+    def match_url(self, url, options=None):
         """
         Return if this rule matches the URL.
 
         What to do if rule is matched is up to developer. Most likely
         ``.is_exception`` attribute should be taken in account.
         """
-        params = params or {}
+        options = options or {}
         for optname in self.options:
             if optname == 'match-case':  # TODO
                 continue
 
-            if optname not in params:
+            if optname not in options:
                 raise ValueError("Rule requires option %s" % optname)
 
             if optname == 'domain':
-                if not self._domain_matches(params['domain']):
+                if not self._domain_matches(options['domain']):
                     return False
                 continue
 
-            if params[optname] != self.options[optname]:
+            if options[optname] != self.options[optname]:
                 return False
 
         return self._url_matches(url)
@@ -139,10 +145,10 @@ class AdblockRule(object):
     def _url_matches(self, url):
         return bool(re.search(self.regex, url))
 
-    def matching_supported(self, params=None):
+    def matching_supported(self, options=None):
         """
         Return whether this rule can return meaningful result,
-        given the `params` dict. If some options are missing,
+        given the `options` dict. If some options are missing,
         then rule shouldn't be matched against, and this function
         returns False.
 
@@ -162,9 +168,9 @@ class AdblockRule(object):
         if self.is_html_rule:  # HTML rules are not supported yet
             return False
 
-        params = params or {}
-        params_keys = set(params.keys())
-        if not params_keys.issuperset(self._options_keys):
+        options = options or {}
+        keys = set(options.keys())
+        if not keys.issuperset(self._options_keys):
             # some of the required options are not given
             return False
 
@@ -210,7 +216,7 @@ class AdblockRule(object):
         # Separator character ^ matches anything but a letter, a digit, or
         # one of the following: _ - . %. The end of the address is also
         # accepted as separator.
-        rule = rule.replace("^", "(?:[^\w\d%_\-.%]|$)")
+        rule = rule.replace("^", "(?:[^\w\d_\-.%]|$)")
 
         # * symbol
         rule = rule.replace("*", ".*")
@@ -268,49 +274,91 @@ class AdblockRules(object):
             if r.regex and r.matching_supported(_params)
         ]
 
-        basic_rules = [r for r in self.rules if not r.options]
-        advanced_rules = [r for r in self.rules if r.options]
+        advanced_rules, basic_rules = split_data(self.rules, lambda r: r.options)
+
+        domain_rules, non_domain_rules = split_data(
+            advanced_rules,
+            lambda r: 'domain' in r.options
+        )
 
         self.blacklist, self.whitelist = self._split_bw(basic_rules)
-        self.blacklist2, self.whitelist2 = self._split_bw(advanced_rules)
-
         _combined = partial(_combined_regex, use_re2=self.uses_re2, max_mem=max_mem)
-
         self.blacklist_re = _combined([r.regex for r in self.blacklist])
         self.whitelist_re = _combined([r.regex for r in self.whitelist])
 
-    def should_block(self, url, params=None):
-        if self.whitelist_re and self.whitelist_re.search(url):
-            return False
+        self.blacklist_adv, self.whitelist_adv = self._split_bw(non_domain_rules)
+        self.blacklist_domains, self.whitelist_domains = self._split_bw_domain(domain_rules)
 
-        params = params or {}
 
+    def should_block(self, url, options=None):
         # TODO: group rules with similar options and match them in bigger steps
+        options = options or {}
+        if self._is_whitelisted(url, options):
+            return False
+        if self._is_blacklisted(url, options):
+            return True
+        return False
 
-        whitelist2 = self.whitelist2
-        blacklist2 = self.blacklist2
-        if self.skip_unsupported_rules:
-            whitelist2 = [rule for rule in self.whitelist2 if rule.matching_supported(params)]
-            blacklist2 = [rule for rule in self.blacklist2 if rule.matching_supported(params)]
+    def _is_whitelisted(self, url, options):
+        return self._matches(
+            url, options,
+            self.whitelist_re, self.whitelist_domains, self.whitelist_adv
+        )
 
-        for rule in whitelist2:
-            if rule.match_url(url, params):
-                return False
+    def _is_blacklisted(self, url, options):
+        return self._matches(
+            url, options,
+            self.blacklist_re, self.blacklist_domains, self.blacklist_adv
+        )
 
-        if self.blacklist_re and self.blacklist_re.search(url):
+    def _matches(self, url, options,
+                 general_re, rules_with_domains, rules_with_options):
+        """
+        Return if ``url``/``options`` are matched by rules defined by
+        ``general_re``, ``rules_with_domains`` and ``rules_with_options``.
+
+        ``general_re`` is a compiled regex for rules without options.
+
+        ``rules_with_domains`` is a {domain: [rules]} mapping - per-domain
+        index of applicable rules which have 'domain' option.
+
+         ``rules_with_options`` is a list of AdblockRule instances that
+        don't have 'domain' option, but have other options.
+        """
+        if general_re and general_re.search(url):
             return True
 
-        for rule in blacklist2:
-            if rule.match_url(url, params):
-                return True
+        rules = []
+        if 'domain' in options and rules_with_domains:
+            src_domain = options['domain']
+            for domain in _domain_variants(src_domain):
+                if domain in rules_with_domains:
+                    rules = rules_with_domains[domain]
+                    break
 
-        return False
+        rules.extend(rules_with_options)
+
+        if self.skip_unsupported_rules:
+            rules = [rule for rule in rules if rule.matching_supported(options)]
+
+        return any(rule.match_url(url, options) for rule in rules)
 
     @classmethod
     def _split_bw(cls, rules):
-        blacklist = [r for r in rules if not r.is_exception]
-        whitelist = [r for r in rules if r.is_exception]
-        return blacklist, whitelist
+        return split_data(rules, lambda r: not r.is_exception)
+
+    @classmethod
+    def _split_bw_domain(cls, rules):
+        blacklist, whitelist = cls._split_bw(rules)
+        return cls._domain_index(blacklist), cls._domain_index(whitelist)
+
+    @classmethod
+    def _domain_index(cls, rules):
+        result = defaultdict(list)
+        for rule in rules:
+            for domain in rule.options.get('domain', []):
+                result[domain].append(rule)
+        return dict(result)
 
 
 def _domain_variants(domain):
@@ -341,11 +389,10 @@ def _combined_regex(regexes, flags=re.IGNORECASE, use_re2=False, max_mem=None):
     if not joined_regexes:
         return None
 
-    regex_str = "(%s)" % joined_regexes
     if use_re2:
         import re2
-        return re2.compile(regex_str, flags=flags, max_mem=max_mem)
-    return re.compile(regex_str, flags=flags)
+        return re2.compile(joined_regexes, flags=flags, max_mem=max_mem)
+    return re.compile(joined_regexes, flags=flags)
 
 
 def _is_re2_supported():
